import sys
import functools
import inspect
import typing as t
from urllib.parse import urlparse as _urlparse
from urllib.parse import urlunparse
from urllib.parse import ParseResult

import click
from flask import Flask
from flask import current_app
from flask import has_app_context
from flask_sqlalchemy import SQLAlchemy

from . import config
from ._compat import psycopg
from .exceptions import SqlaExtensionNotFound
from .exceptions import EnvironmentNotAllowed


@functools.wraps(_urlparse)
def urlparse(uri: str, *args, **kwargs) -> ParseResult:
    """
    Parse database URLs + builtin validation. Better to validate here than to
    stumble into an unpredictable and cryptic error message down the line.
    """
    data = _urlparse(uri, *args, **kwargs)
    assert data.scheme, (
        f"The URI {uri!r} does not have a valid scheme."
    )
    assert data.netloc, (
        f"The URI {uri!r} does not have a valid host."
    )
    assert data.path, (
        f"The URI {uri!r} does not have a valid path/database name."
    )
    return data


@functools.wraps(click.echo)
def echo(*args, **kwargs) -> None:
    """
    Most of our outputs relate to significant changes to the state of a file
    system and/or a database, so we should print to `stderr` by default.
    """
    kwargs.setdefault("file", sys.stderr)
    return click.echo(*args, **kwargs)


def swap_db_name(
        uri: str,
        new_db_name: str
) -> str:
    data = urlparse(uri)._asdict()
    data["path"] = new_db_name
    return urlunparse(ParseResult(**data))


def extract_db_name(uri: str) -> str:
    return urlparse(uri).path.replace("/", "", 1)


def get_admin_uri(
        target_uri: t.Optional[str] = None,
        admin_path: t.Optional[str] = None
) -> str:
    def _parse_admin_path(s: str) -> str:
        if s.startswith("/"):
            s = s[1:]
        return s

    return swap_db_name(
        target_uri or config.get("FLASK_POSTGRES_TARGET_DATABASE_URI"),
        admin_path or _parse_admin_path(config.get("FLASK_POSTGRES_ADMIN_PATH")),
    )


def get_db() -> SQLAlchemy:
    if not has_app_context():
        # RuntimeError is the same err type raised by flask.globals
        raise RuntimeError("The app context is currently not active.")
    if "sqlalchemy" not in current_app.extensions:
        raise SqlaExtensionNotFound
    return current_app.extensions["sqlalchemy"].db


def raise_err_if_disallowed():
    """
    You can protect your app against accidental or (some, but not much)
    malicious use in sensitive environments with the
    `FLASK_POSTGRES_CLI_DISALLOWED_ENVS` config variable.
    """
    li = config.get("FLASK_POSTGRES_CLI_DISALLOWED_ENVS")
    if isinstance(li, str):
        li = li.split(";")
    if current_app.env in li:
        raise EnvironmentNotAllowed


def database_exists(
        database_name: str,
        conn: psycopg.connection.Connection
) -> bool:
    cursor = conn.cursor()

    try:
        exists = bool(cursor.execute(
            "SELECT MAX((datname = %s)::int) FROM pg_database;",
            (database_name,)
        ).fetchone()[0])
    finally:
        cursor.close()

    return exists


def resolve_init_db_callback(
        callback: callable
) -> t.Callable[[Flask, SQLAlchemy], t.Any]:

    app_in_params = "app" in inspect.signature(callback).parameters
    db_in_params = "db" in inspect.signature(callback).parameters

    # code is a little wet, but I don't mind avoiding nonlocals and over-the-
    # top abstraction.

    if app_in_params and db_in_params:
        def resolved_callback(app: Flask, db: SQLAlchemy):
            return callback(app=current_app, db=get_db())
    elif app_in_params and not db_in_params:
        def resolved_callback(app: Flask, db: SQLAlchemy):
            return callback(app=current_app)
    elif not app_in_params and db_in_params:
        def resolved_callback(app: Flask, db: SQLAlchemy):
            return callback(db=get_db())
    else:
        def resolved_callback(app: Flask, db: SQLAlchemy):
            return callback()

    return resolved_callback


def echo_error_as_warning(e: Exception) -> None:
    msg = e.args[0]
    if "FATAL:" in msg:  # psycopg2 exceptions
        msg = ":".join(msg.split(":")[1:]).strip()
    echo(click.style(msg, fg="yellow"), )



import pytest
import os
from pytest_postgresql.janitor import DatabaseJanitor
import sqlalchemy as sa

from flask import Flask
from flask_postgres._compat import psycopg  # noqa
from flask_postgres.cli import cli


@pytest.fixture
def uri() -> str:
    return os.environ["TEST_DATABASE_URI"]


@pytest.fixture
def db_params(uri) -> dict:
    data = sa.engine.url.make_url(uri).translate_connect_args()
    data.setdefault("dbname", data.pop("database", "localhost"))
    data.setdefault("user", data.pop("username", None))
    data.setdefault("host", data.pop("host", 5432))
    return data


@pytest.fixture(autouse=True)
def clean_env(monkeypatch):
    monkeypatch.delitem(os.environ, "FLASK_POSTGRES_TARGET_DATABASE_URI", raising=False)
    monkeypatch.delitem(os.environ, "FLASK_POSTGRES_ADMIN_PATH", raising=False)
    monkeypatch.delitem(os.environ, "SQLALCHEMY_DATABASE_URI", raising=False)
    monkeypatch.delitem(os.environ, "FLASK_POSTGRES_CLI_DISALLOWED_ENVS", raising=False)
    yield


@pytest.fixture
def base_app() -> Flask:
    app = Flask("test")
    app.config["ENV"] = "testing"
    app.config["TESTING"] = True
    app.config["DEBUG"] = True
    app.cli.add_command(cli)
    return app


@pytest.fixture(autouse=True)
def admin_connection(db_params) -> psycopg.connection.Connection:
    admin_params = db_params.copy()
    admin_params["dbname"] = "postgres"
    conn = psycopg.connect(**admin_params)
    conn.autocommit = True
    yield conn
    conn.close()


@pytest.fixture(autouse=True)
def uninitialized_database_janitor(db_params: dict):
    """Set up an environment in which the database specified in db_params is
    always cleaned up, even if it wasn't created.

    Creating the database is handled either by other fixtures, or test cases
    themselves.
    """
    janitor = DatabaseJanitor(**db_params, version=9.6)
    yield janitor

    try:
        # Always attempt to drop the database!
        janitor.drop()
    except psycopg.errors.InvalidCatalogName:
        pass


@pytest.fixture
def activated_database(
        db_params: dict,
        uninitialized_database_janitor: DatabaseJanitor
):
    """Initializes the `uninitialized_database_janitor` object.

    We do not return a psycopg.connect() to it because that makes it impossible
    to delete, and some of our tests need to cover the deletion case.
    """
    uninitialized_database_janitor.init()
    yield



import os
import inspect
import pytest
import typing as t

from flask_sqlalchemy import SQLAlchemy

from flask_postgres.utils import swap_db_name
from flask_postgres.utils import database_exists
from flask_postgres.utils import resolve_init_db_callback
from flask_postgres.utils import get_db
from flask_postgres.utils import raise_err_if_disallowed
from flask_postgres.utils import get_admin_uri
from flask_postgres.exceptions import SqlaExtensionNotFound
from flask_postgres.exceptions import EnvironmentNotAllowed
from flask_postgres.exceptions import PostgresUrlNotSet


def test_swap_db_name_function():
    res = swap_db_name(
        "postgres://foo@bar:hello.world/replace_me",
        "replaced_with"
    )
    assert res == "postgres://foo@bar:hello.world/replaced_with"


def test_database_exists_function(
        admin_connection,
        db_params
):
    """It's critical that this test passes because other tests use the
    `database_exists` function to run their own tests.
    """
    dbname = db_params.get("dbname")
    assert not database_exists(database_name=dbname, conn=admin_connection)

    cursor = admin_connection.cursor()
    cursor.execute(f"CREATE DATABASE {dbname};")
    assert database_exists(database_name=dbname, conn=admin_connection)

    cursor.execute(f"DROP DATABASE {dbname};")
    assert not database_exists(database_name=dbname, conn=admin_connection)


@pytest.mark.parametrize(
    ["func", "app_is_arg", "db_is_arg"], [
        (lambda: print("I ran OK"), False, False),
        (lambda app: print("I ran OK"), True, False),
        (lambda db: print("I ran OK"), False, True),
        (lambda app, db: print("I ran OK"), True, True),
        (lambda db, app: print("I ran OK"), True, True),
    ])
@pytest.mark.filterwarnings("ignore:.*SQLALCHEMY.*")
def test_resolve_init_db_callback_function(
    base_app,
    func: callable,
    app_is_arg: bool,
    db_is_arg: bool,
capsys
):
    assert app_is_arg == bool("app" in inspect.signature(func).parameters)
    assert db_is_arg == bool("db" in inspect.signature(func).parameters)

    new_func = resolve_init_db_callback(func)

    assert "app" in inspect.signature(new_func).parameters
    assert "db" in inspect.signature(new_func).parameters

    db = SQLAlchemy(base_app)

    assert capsys.readouterr().out == ""

    with base_app.app_context():
        new_func(app=base_app, db=db)  # noqa

    assert capsys.readouterr().out == "I ran OK\n"


@pytest.mark.filterwarnings("ignore:.*SQLALCHEMY.*")
def test_get_db_function(base_app):
    # No db extension
    with pytest.raises(SqlaExtensionNotFound):
        with base_app.app_context():
            get_db()

    # This works perfectly fine.
    db1 = SQLAlchemy(base_app)
    with base_app.app_context():
        db2 = get_db()
    assert db1 is db2

    # No app context
    with pytest.raises(RuntimeError):
        get_db()


def test_raise_err_if_disallowed_function(base_app):
    # Nothing should happen
    with base_app.app_context():
        raise_err_if_disallowed()

    base_app.config["FLASK_POSTGRES_CLI_DISALLOWED_ENVS"] = [base_app.env]

    # Now we should see an error
    with pytest.raises(EnvironmentNotAllowed):
        with base_app.app_context():
            raise_err_if_disallowed()


class AdminUriTestConf(t.NamedTuple):
    app_config: t.Dict[str, t.Any]
    expected_admin_uri: str
    env_vars: t.Optional[t.Dict[str, str]] = None


typical_config = AdminUriTestConf(
    app_config={
        "SQLALCHEMY_DATABASE_URI": "postgresql://foo:bar@hello:12345/world"
    },
    expected_admin_uri="postgresql://foo:bar@hello:12345/postgres"
)

typical_config_w_custom_admin = AdminUriTestConf(
    app_config={
        "SQLALCHEMY_DATABASE_URI": "postgresql://foo:bar@hello:12345/world",
        "FLASK_POSTGRES_ADMIN_PATH": "johndoe"
    },
    expected_admin_uri="postgresql://foo:bar@hello:12345/johndoe"
)

typical_config_w_custom_admin_w_forward_slash = AdminUriTestConf(
    app_config={
        "SQLALCHEMY_DATABASE_URI": "postgresql://foo:bar@hello:12345/world",
        "FLASK_POSTGRES_ADMIN_PATH": "/johndoe"
    },
    expected_admin_uri="postgresql://foo:bar@hello:12345/johndoe"
)

config_with_env_var = AdminUriTestConf(
    app_config={
        "SQLALCHEMY_DATABASE_URI": "postgresql://foo:bar@hello:12345/world",
    },
    expected_admin_uri="postgresql://foo:bar@hello:12345/johndoe",
    env_vars={"FLASK_POSTGRES_ADMIN_PATH": "johndoe"}
)

config_with_env_var_that_gets_overruled = AdminUriTestConf(
    app_config={
        "SQLALCHEMY_DATABASE_URI": "postgresql://foo:bar@hello:12345/world",
        "FLASK_POSTGRES_ADMIN_PATH": "johndoe"
    },
    expected_admin_uri="postgresql://foo:bar@hello:12345/johndoe",
    env_vars={"FLASK_POSTGRES_ADMIN_PATH": "overrule_me_please"}
)

config_with_sqla_uri_that_gets_overruled = AdminUriTestConf(
    app_config={
        "SQLALCHEMY_DATABASE_URI": "postgresql://dont:use@me:12345/please",
        "FLASK_POSTGRES_TARGET_DATABASE_URI": "postgresql://foo:bar@hello:12345/world",
        "FLASK_POSTGRES_ADMIN_PATH": "johndoe"
    },
    expected_admin_uri="postgresql://foo:bar@hello:12345/johndoe",
)

config_with_env_var_that_overrules_sqla_uri = AdminUriTestConf(
    app_config={
        "SQLALCHEMY_DATABASE_URI": "postgresql://dont:use@me:12345/please",
        "FLASK_POSTGRES_ADMIN_PATH": "johndoe"
    },
    expected_admin_uri="postgresql://foo:bar@hello:12345/johndoe",
    env_vars={
        "FLASK_POSTGRES_TARGET_DATABASE_URI": "postgresql://foo:bar@hello:12345/world"
    }
)


@pytest.mark.parametrize(
    "conf", [
        typical_config,
        typical_config_w_custom_admin,
        typical_config_w_custom_admin_w_forward_slash,
        config_with_env_var,
        config_with_env_var_that_gets_overruled,
        config_with_sqla_uri_that_gets_overruled,
        config_with_env_var_that_overrules_sqla_uri
    ]
)
def test_get_admin_uri_function(base_app, conf, monkeypatch):
    """This app critically depends on how it resolves config variables. So this
    test doesn't just cover the get_admin_uri function, it also covers all of
    the config resolution logic.
    """
    if conf.env_vars:
        for k, v in conf.env_vars.items():
            monkeypatch.setitem(os.environ, k, v)

    base_app.config.update(conf.app_config)

    with base_app.app_context():
        res = get_admin_uri()

    assert res == conf.expected_admin_uri


def test_get_admin_uri_raises_err_if_no_config(base_app):

    with pytest.raises(PostgresUrlNotSet):
        with base_app.app_context():
            get_admin_uri()



import sys
import contextlib
import typing as t

import click
import sqlalchemy.exc
from flask import current_app
from flask.cli import AppGroup
from flask.cli import ScriptInfo
from flask.cli import with_appcontext

from . import config
from ._compat import check_dependencies
from ._compat import psycopg
from .utils import extract_db_name
from .utils import get_admin_uri
from .utils import get_db
from .utils import echo_error_as_warning
from .utils import echo
from .utils import raise_err_if_disallowed
from .utils import resolve_init_db_callback
from .utils import database_exists

FC = t.TypeVar("FC", t.Callable[..., t.Any], click.Command)


pass_psycopg_connection = \
    click.make_pass_decorator(psycopg.connection.Connection)


@contextlib.contextmanager
def may_not_exist_context():
    try:
        yield
    except (
            psycopg.errors.OperationalError,
            sqlalchemy.exc.OperationalError,
    ) as e:
        if e.code != "e3q8":
            raise
        echo_error_as_warning(e)
    except psycopg.errors.InvalidCatalogName as e:
        echo_error_as_warning(e)


@contextlib.contextmanager
def may_already_exist_context():
    try:
        yield
    except psycopg.errors.DuplicateDatabase as e:
        echo_error_as_warning(e)


class ContextMixin(click.Parameter):

    def get_default(
            self,
            ctx: click.Context,
            *args,
            **kwargs
    ) -> t.Optional[t.Any]:
        with ctx.ensure_object(ScriptInfo).load_app().app_context():
            return super().get_default(ctx, *args, **kwargs)


class ContextOption(ContextMixin, click.Option):
    """Option that can load defaults with an active app context."""
    pass


class ContextArgument(ContextMixin, click.Argument):
    """Argument that can load defaults with an active app context."""
    pass


def get_connection(ctx: click.Context, admin_uri: str) -> psycopg.Connection:
    conn = psycopg.connect(admin_uri)
    conn.autocommit = True
    ctx.call_on_close(conn.close)
    return conn


@click.group("psql", cls=AppGroup)
@with_appcontext
def cli():
    """
    Manage your PostgreSQL database instance.
    """
    check_dependencies()
    raise_err_if_disallowed()


cli: AppGroup


def database_name_option(
        operation_past_tense: str
) -> t.Callable[[FC], FC]:
    return click.option(
        "--database-name", "-d",
        cls=ContextOption,
        default=lambda: extract_db_name(
            config.get("FLASK_POSTGRES_TARGET_DATABASE_URI")),
        show_default=True,
        help=f"The URI of the database to be {operation_past_tense}. By"
             " default, this is derived from the SQLALCHEMY_DATABASE_URI."
    )


admin_uri_option = click.option(
    "--admin-uri", "-u", "admin_uri",
    cls=ContextOption,
    default=get_admin_uri,
    show_default=True,
    help="PostgreSQL database URI to create. By default, this is your"
         " SQLALCHEMY_DATABASE_URI but with 'postgres' as the path."
)


force_delete_option = click.option(
    "--force", "-f",
    cls=ContextOption,
    is_flag=True,
    default=False,
    help="If true, then force delete the database."
)


@cli.command("create")
@database_name_option("created")
@admin_uri_option
@with_appcontext
@click.pass_context
def create_db_command(
        ctx: click.Context,
        database_name: str,
        admin_uri: str,
        **kwargs
):
    """Create the database."""
    conn = get_connection(ctx, admin_uri)
    cursor = conn.cursor()
    try:
        with may_already_exist_context():
            cursor.execute(f"CREATE DATABASE {database_name};")
            echo(f'database "{database_name}" was created')
    finally:
        cursor.close()


create_db_command: click.Command


@cli.command("init")
@with_appcontext
def init_db_command(**kwargs):
    """Init the database."""
    with may_not_exist_context():
        db = get_db()
        if config.registered_callback:
            callback = resolve_init_db_callback(config.registered_callback)
            callback(current_app, db)
        else:
            db.create_all()
        database_name = kwargs.get("database_name") \
            or extract_db_name(str(db.engine.url))
        echo(f'database "{database_name}" was initialized')


init_db_command: click.Command


@cli.command("delete")
@database_name_option("deleted")
@admin_uri_option
@force_delete_option
@with_appcontext
@click.pass_context
def delete_db_command(
        ctx: click.Context,
        database_name: str,
        admin_uri: str,
        force: bool,
        **kwargs
):
    """Delete the database."""
    conn = get_connection(ctx, admin_uri)
    if force is not True:

        if not database_exists(database_name=database_name, conn=conn):
            msg = f'database "{database_name}" does not exist'
            echo(click.style(msg, fg="yellow"))
            return

        # Don't print the prompt to stderr
        # So use click.echo() instead of echo().
        click.echo(
            f'Are you sure you want to delete database "{database_name}"?'
        )
        click.echo(
            "Type in the database name"
            f' "{click.style(database_name, fg="red")}"'
            " to confirm.")
        confirmation = click.prompt("Database name")

        if database_name != confirmation:
            echo(click.style(
                f'The input "{confirmation}" does not match the database name.'
                f' The database "{database_name}" will not be deleted.',
                fg="red"
            ))
            sys.exit(1)

    cursor = conn.cursor()

    try:
        with may_not_exist_context():
            cursor.execute(f"DROP DATABASE {database_name};")
            echo(f'database "{database_name}" was deleted')
    finally:
        cursor.close()


delete_db_command: click.Command


@cli.command("setup")
@database_name_option("setup")
@admin_uri_option
@click.pass_context
@with_appcontext
def setup_db_command(
        ctx: click.Context,
        **kwargs
):
    """Create and init the database."""
    ctx.forward(create_db_command)
    ctx.forward(init_db_command)


setup_db_command: click.Command


@cli.command("reset")
@database_name_option("reset")
@admin_uri_option
@force_delete_option
@click.pass_context
@with_appcontext
def reset_db_command(
        ctx: click.Context,
        **kwargs
):
    """Delete, then create and init the database."""
    ctx.forward(delete_db_command)
    ctx.forward(create_db_command)
    ctx.forward(init_db_command)


reset_db_command: click.Command



import os
import typing as t
import warnings

from flask import has_app_context
from flask import current_app

from .exceptions import PostgresUrlNotSet


T = t.TypeVar("T")


# Lazy Loaders


def _target_database_uri_default() -> str:
    db_path = get("SQLALCHEMY_DATABASE_URI")
    if db_path:
        return db_path
    else:
        raise PostgresUrlNotSet


DEFAULT_CONFIG: t.Dict[str, t.Callable[[], t.Any]] = {
    "FLASK_POSTGRES_CLI_DISALLOWED_ENVS": lambda: set(),
    "FLASK_POSTGRES_ADMIN_PATH": lambda: "postgres",
    "FLASK_POSTGRES_TARGET_DATABASE_URI": _target_database_uri_default,
}


def get(key: str) -> t.Any:
    if has_app_context() and key in current_app.config:
        return current_app.config.get(key)
    elif key in os.environ:
        return os.environ.get(key)
    elif key in DEFAULT_CONFIG:
        return DEFAULT_CONFIG.get(key)()
    else:
        return None


def init_db_callback(func: T) -> T:
    current_callback = globals().get("registered_callback")
    if current_callback:
        warnings.warn(
            "One or more init_db_callback's were registered, even though"
            f" a init_db_callback, {current_callback!r}, already exists."
            f" The new callback, {func}, will be registered and replace the"
            " existing one.",
            UserWarning
        )
    globals()["registered_callback"] = func
    return func


def reset_init_db_callback():
    globals()["registered_callback"] = None


registered_callback: callable = None



class FlaskPostgresException(Exception):

    default: str = None

    def __init__(self, *args):
        if len(args) == 0 and self.default:
            args = [self.default]
        super().__init__(*args)


class SqlaExtensionNotFound(
        FlaskPostgresException,
        AttributeError
):
    default = (
        "The sqlalchemy extension was not registered to the current"
        " application."
    )


class EnvironmentNotAllowed(
        FlaskPostgresException,
        EnvironmentError
):
    default = (
        "This app has been configured so that Flask-Postgres does not run in"
        " this environment."
    )


class PostgresUrlNotSet(
        FlaskPostgresException,
        ValueError
):
    default = (
        "A target database URI has not been set. Please either set up the"
        " SQLALCHEMY_DATABASE_URI config variable (recommended), or"
        " alternatively, designate a database URI via the"
        " FLASK_POSTGRES_TARGET_DATABASE_URI config variable."
    )



"""
Compatibility for psycopg.

This repo does not enforce psycopg in the requirements because users may be
using psycopg2 or the psycopg2-binary. All of these installations have a very
similar API. Compatibility across versions is handled here.
"""
_MODULE_NOT_FOUND_MSG = (
    "No module named 'psycopg' or 'psycopg2'."
    " Flask-Postgres depends on either of these modules to handle connections"
    " to PostgreSQL."
)

try:
    import psycopg as psycopg
except ModuleNotFoundError:
    try:
        import psycopg2 as psycopg
    except ModuleNotFoundError:
        psycopg = None


def check_dependencies():
    """
    At some point we can't beat around the bush, and we need to raise an error
    if there is no psycopg module.
    """
    if psycopg is None:
        raise ModuleNotFoundError(_MODULE_NOT_FOUND_MSG)



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="Flask-Postgres",
    install_requires=[
        "Flask>=1.0",
        "Click>=6.0",
        "SQLAlchemy>=1.2.2",
        "Flask-SQLAlchemy>=2.4",
    ],
    extras_require={
        "tests": [
            "psycopg",
            "pytest>=6.0.1",
            "pytest-env",
            "pytest-cov",
            "pytest-postgresql"
        ]
    },
)



import typing as t
import pytest
import sqlalchemy as sa
from flask import Flask
from flask_sqlalchemy import SQLAlchemy

from flask_postgres.utils import database_exists
from flask_postgres import init_db_callback
from flask_postgres.config import reset_init_db_callback


@pytest.fixture
def db() -> SQLAlchemy:
    return SQLAlchemy()


@pytest.fixture
def example_model(db):
    class Pet(db.Model):
        id = db.Column(db.Integer, primary_key=True, autoincrement=True)
        name = db.Column(db.Text)
    return Pet


@pytest.fixture
def typically_configured_app(base_app, uri, db, example_model) -> Flask:
    base_app.config["SQLALCHEMY_DATABASE_URI"] = uri
    base_app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
    db.init_app(base_app)
    return base_app


@pytest.fixture
def custom_init_db_callback(
        typically_configured_app: Flask,
        db: SQLAlchemy,
        example_model
):

    @init_db_callback
    def callback(db: SQLAlchemy):
        db.create_all()
        pet = example_model(name="Fido")
        db.session.add(pet)
        db.session.commit()

    yield

    reset_init_db_callback()


def test_psql_create_command(
        admin_connection,
        typically_configured_app: Flask,
        db_params: dict
):
    dbname = db_params.get("dbname")

    assert not database_exists(dbname, admin_connection)

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(typically_configured_app.cli, ["psql", "create"])

    assert res.exit_code == 0
    assert res.output == f'database "{dbname}" was created\n'
    assert database_exists(dbname, admin_connection)


def test_psql_init_command(
        typically_configured_app: Flask,
        activated_database,
        db,
        example_model,
        db_params: dict
):
    dbname = db_params.get("dbname")

    with typically_configured_app.app_context():
        db_inspector = sa.inspect(db.engine)
        assert not db_inspector.has_table("pet")

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(typically_configured_app.cli, ["psql", "init"])

    assert res.exit_code == 0
    assert res.output == f'database "{dbname}" was initialized\n'

    with typically_configured_app.app_context():
        db_inspector = sa.inspect(db.engine)
        assert db_inspector.has_table("pet")

    with typically_configured_app.app_context():
        first_row_of_data = db.session.query(example_model).first()
    assert first_row_of_data is None


def test_psql_init_command_when_db_not_created(
        typically_configured_app: Flask,
        db,
        example_model,
        db_params: dict,
        admin_connection
):
    dbname = db_params.get("dbname")

    assert not database_exists(dbname, admin_connection)

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(typically_configured_app.cli, ["psql", "init"])

    assert res.output == f'database "{dbname}" does not exist\n'
    assert not database_exists(dbname, admin_connection)


def test_psql_init_command_with_custom_callback(
        typically_configured_app: Flask,
        activated_database,
        db_params: dict,
        db,
        example_model,
        custom_init_db_callback,
):
    dbname = db_params.get("dbname")

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(typically_configured_app.cli, ["psql", "init"])

    assert res.exit_code == 0
    assert res.output == f'database "{dbname}" was initialized\n'

    with typically_configured_app.app_context():
        first_row_of_data = db.session.query(example_model).first()

    assert first_row_of_data is not None


class CliTestConf(t.NamedTuple):
    options: t.List[str]
    type_in_db_name: bool = False


@pytest.mark.parametrize(
    "cli_test_conf", [
        CliTestConf(options=[], type_in_db_name=True),
        CliTestConf(options=["-f"]),
        CliTestConf(options=["--force"]),
        CliTestConf(options=["--force"]),
    ]
)
def test_psql_delete_command(
        admin_connection,
        typically_configured_app: Flask,
        activated_database,
        db_params: dict,
        cli_test_conf: CliTestConf
):
    dbname = db_params.get("dbname")

    assert database_exists(dbname, admin_connection)

    if cli_test_conf.type_in_db_name:
        input_stream = dbname
    else:
        input_stream = None

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(
        typically_configured_app.cli,
        ["psql", "delete", *cli_test_conf.options],
        input=input_stream
    )

    assert res.exit_code == 0

    if cli_test_conf.type_in_db_name:
        assert res.output.startswith("Are you sure")
        assert res.output.endswith(f'\ndatabase "{dbname}" was deleted\n')
    else:
        assert res.output == f'database "{dbname}" was deleted\n'

    assert not database_exists(dbname, admin_connection)


def test_psql_delete_command_wont_delete_on_typo(
        admin_connection,
        typically_configured_app: Flask,
        activated_database,
        db_params: dict,
):
    dbname = db_params.get("dbname")

    assert database_exists(dbname, admin_connection)

    input_stream = dbname + "extratext"

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(
        typically_configured_app.cli,
        ["psql", "delete"],
        input=input_stream
    )

    assert res.exit_code == 1

    assert res.output.startswith("Are you sure")
    assert f'\ndatabase "{dbname}" was deleted\n' not in res.output
    assert f'database "{dbname}" will not be deleted' in res.output

    assert database_exists(dbname, admin_connection)


def test_psql_setup_command(
        admin_connection,
        typically_configured_app: Flask,
        db_params: dict,
        db,
        example_model
):
    dbname = db_params.get("dbname")

    assert not database_exists(dbname, admin_connection)

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(typically_configured_app.cli, ["psql", "setup"])

    assert res.exit_code == 0
    assert res.output == (
        f'database "{dbname}" was created\n'
        f'database "{dbname}" was initialized\n'
    )

    assert database_exists(dbname, admin_connection)

    with typically_configured_app.app_context():
        first_row_of_data = db.session.query(example_model).first()
    assert first_row_of_data is None


def test_psql_setup_command_with_custom_callback(
        admin_connection,
        typically_configured_app: Flask,
        db_params: dict,
        db,
        example_model,
        custom_init_db_callback,
):
    dbname = db_params.get("dbname")

    assert not database_exists(dbname, admin_connection)

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(typically_configured_app.cli, ["psql", "setup"])

    assert res.exit_code == 0
    assert res.output == (
        f'database "{dbname}" was created\n'
        f'database "{dbname}" was initialized\n'
    )

    assert database_exists(dbname, admin_connection)

    with typically_configured_app.app_context():
        first_row_of_data = db.session.query(example_model).first()
    assert first_row_of_data is not None


def test_psql_setup_command_when_db_exists(
        admin_connection,
        activated_database,
        typically_configured_app: Flask,
        db_params: dict,
        db,
        example_model
):
    dbname = db_params.get("dbname")

    assert database_exists(dbname, admin_connection)

    with typically_configured_app.app_context():
        db_inspector = sa.inspect(db.engine)
        assert not db_inspector.has_table("pet")

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(typically_configured_app.cli, ["psql", "setup"])

    assert res.exit_code == 0
    assert res.output == (
        f'database "{dbname}" already exists\n'
        f'database "{dbname}" was initialized\n'
    )

    assert database_exists(dbname, admin_connection)

    with typically_configured_app.app_context():
        db_inspector = sa.inspect(db.engine)
        assert db_inspector.has_table("pet")

    with typically_configured_app.app_context():
        first_row_of_data = db.session.query(example_model).first()
    assert first_row_of_data is None


def test_psql_reset_command(
        admin_connection,
        typically_configured_app: Flask,
        activated_database,
        db_params: dict,
        db,
        example_model
):
    dbname = db_params.get("dbname")

    assert database_exists(dbname, admin_connection)

    client = typically_configured_app.test_cli_runner()
    res = client.invoke(typically_configured_app.cli, ["psql", "reset", "-f"])

    assert res.exit_code == 0
    assert res.output == (
        f'database "{dbname}" was deleted\n'
        f'database "{dbname}" was created\n'
        f'database "{dbname}" was initialized\n'
    )

    assert database_exists(dbname, admin_connection)

    with typically_configured_app.app_context():
        db_inspector = sa.inspect(db.engine)
        assert db_inspector.has_table("pet")

    with typically_configured_app.app_context():
        first_row_of_data = db.session.query(example_model).first()
    assert first_row_of_data is None
